"""This module implements the StateManager class."""

__all__ = ['StateManager']

from collections.abc import Iterable
from typing import TypeVar, overload

from pyriak import _SENTINEL, EventQueue
from pyriak.events import StateAdded, StateRemoved


_T = TypeVar('_T')
_D = TypeVar('_D')


class StateManager:
  """A manager and container of states.

  All objects are valid states.

  The StateManager can be thought of as the program's dedicated Entity
  instance, where global and unique data can be put, e.g. global time.
  A common use is a system storing data in a state that it 'owns',
  as systems are not allowed to have any mutable data themselves.

  The general purpose of the StateManager is to provide storage for data
  that does not make sense to have multiple of, nor put in an entity.

  The StateManager is owned directly by the Space.
  There is usually only one, accessible via the Space's 'states' attribute.

  In many aspects, StateManager is identical to Entity.
  Like Entity, the structure of StateManager is akin to a set of states mixed
  with a dict of state types to states. States don't need to be hashable.

  Attributes:
    event_queue: The optional event queue that the StateManager may post to.
      This is usually assigned the space's event queue.
  """

  __slots__ = '_states', 'event_queue', '__weakref__'

  def __init__(
    self, states: Iterable[object] = (), /, event_queue: EventQueue | None = None
  ):
    """Initialize the StateManager with the given states and event queue.

    By default, the StateManager is initialized with no states and
    event queue as None.

    Args:
      states: The iterable of initial states. Defaults to no states.
      event_queue: The event queue to post to. Defaults to None.
    """
    self.event_queue = event_queue
    self._states: dict[type, object] = {}
    self.add(*states)

  def add(self, *states: object) -> None:
    """Add an arbitrary number of states to self.

    Each state is stored by its class.

    If self already has a state of the same type, an exception
    is raised, preventing the rest of the states from being added
    but not affecting the ones already added.

    If self's event queue is not None, a StateAdded event is
    posted for each state added.

    Args:
      *states: The states to be added.

    Raises:
      ValueError: If self already has a state of the same type.
    """
    self_states = self._states
    event_queue = self.event_queue
    for state in states:
      state_type = type(state)
      if state_type in self_states:
        raise ValueError(f'state manager already has state of type {state_type}')
      self_states[state_type] = state
      if event_queue is not None:
        event_queue.append(StateAdded(state))

  def update(self, *states: object) -> None:
    """Update self with an arbitrary number of states.

    Each state is stored by its class.

    If self already has an existing state of the exact same type, that
    existing state is removed right before adding the provided state.
    This is unless the two states are equivalent or the same object,
    in which case the provided state is skipped without posting events.

    If self's event queue is not None, each state added and removed
    generates a StateAdded and StateRemoved event, respectively.

    Args:
      *states: The states to update self with.
    """
    self_states = self._states
    event_queue = self.event_queue
    for state in states:
      state_type = type(state)
      if state_type in self_states:
        other_state = self_states[state_type]
        if other_state is state or other_state == state:
          continue
        if event_queue is not None:
          event_queue.append(StateRemoved(other_state))
      self_states[state_type] = state
      if event_queue is not None:
        event_queue.append(StateAdded(state))

  def remove(self, *states: object) -> None:
    """Remove an arbitrary number of states from self.

    For each provided state, an existing state in self of
    the exact same type is removed. The existing state must
    be equivalent or the same object as the provided state.

    If no such state is found in self, an exception is raised,
    preventing the rest of the states from being removed
    but not affecting the states already removed.

    If self's event queue is not None, a StateRemoved event is
    posted for each state removed.

    Args:
      *states: The states to be removed.

    Raises:
      ValueError: If one of the states is not found in self.
    """
    self_states = self._states
    event_queue = self.event_queue
    for state in states:
      state_type = type(state)
      try:
        other_state = self_states[state_type]
      except KeyError:
        pass
      else:
        if other_state is state or other_state == state:
          del self_states[state_type]
          if event_queue is not None:
            event_queue.append(StateRemoved(other_state))
          continue
      raise ValueError(state)

  def __getitem__(self, state_type: type[_T], /) -> _T:
    return self._states[state_type]  # type: ignore[return-value]

  def __setitem__(self, state_type: type[_T], state: _T):
    if type(state) is not state_type:
      raise TypeError(state)
    self.update(state)

  def __delitem__(self, state_type: type, /):
    self.remove(self[state_type])

  @overload
  def get(self, state_type: type[_T], /) -> _T | None: ...
  @overload
  def get(self, state_type: type[_T], default: _D, /) -> _T | _D: ...
  def get(self, state_type, default=None, /):
    return self._states.get(state_type, default)

  @overload
  def pop(self, state_type: type[_T], /) -> _T: ...
  @overload
  def pop(self, state_type: type[_T], default: _D, /) -> _T | _D: ...
  def pop(self, state_type, default=_SENTINEL, /):
    try:
      state = self[state_type]
    except KeyError:
      if default is _SENTINEL:
        raise
      return default
    self.remove(state)
    return state

  def types(self):
    return self._states.keys()

  def __iter__(self):
    return iter(self._states.values())

  def __reversed__(self):
    return reversed(self._states.values())

  def __len__(self):
    return len(self._states)

  def __contains__(self, obj: object, /):
    return obj in self._states

  def clear(self) -> None:
    self.remove(*self)



"""This module implements the Entity class."""

__all__ = ['Entity', 'EntityId']

from collections.abc import Iterable
from typing import TYPE_CHECKING, NewType, TypeVar, overload
from uuid import uuid4

from pyriak import _SENTINEL, dead_weakref


if TYPE_CHECKING:
  from weakref import ref as weakref

  from pyriak.entity_manager import EntityManager


EntityId = NewType('EntityId', int)
"""Unique identifier for entities."""


_T = TypeVar('_T')
_D = TypeVar('_D')


class Entity:
  """A mutable collection of components that represents an entity.

  All objects are valid components.

  Entities do not have any behavior on their own. An entity's
  components define its data and systems define their behavior.
  The Entity class should rarely be subclassed because
  functionality is not implemented directly on the Entity class.

  The general purpose of entities is to group together the components
  that make up an object and provide structure to the Space's data.

  Entities are stored in the Space's EntityManager, where they and
  their components are operated on in bulk by the Space's systems.

  An entity's structure is analagous to a set of components
  mixed with a dict of component types to components.
  Components do not need to be hashable.

  Attributes:
    id: A unique integer identifier generated when the entity is created.
      This id can be used to weakly reference and store the entity.
  """

  __slots__ = 'id', '_components', '_manager'

  def __init__(self, components: Iterable[object] = (), /):
    """Initialize the entity with the given components.

    The entity also gets a unique EntityId.
    By default, the entity is initialized with no components.

    Args:
      components: The iterable of initial components. Defaults to no components.
    """
    self.id: EntityId = self.new_id()
    self._manager: weakref[EntityManager] = dead_weakref
    comp_dict: dict[type, object] = {}
    for comp in components:
      comp_type = type(comp)
      if comp_type in comp_dict and (
          (other := comp_dict[comp_type]) is comp or other == comp):
        continue
      comp_dict[comp_type] = comp
    self._components: dict[type, object] = comp_dict

  def add(self, *components: object) -> None:
    """Add an arbitrary number of components to self.

    Each component is stored by its class.

    If self already has a component of the same type, an exception
    is raised, preventing the rest of the components from being added
    but not affecting the ones already added.

    If self is in an EntityManager, a ComponentAdded event is
    posted in the manager event queue for each component added.

    Args:
      *components: The components to be added.

    Raises:
      ValueError: If self already has a component of the same type.
    """
    self_components = self._components
    manager = self._manager()
    for component in components:
      component_type = type(component)
      if component_type in self_components:
        raise ValueError(f'entity already has component of type {component_type}')
      self_components[component_type] = component
      if manager is not None:
        manager._component_added(self, component)

  def update(self, *components: object) -> None:
    """Update self with an arbitrary number of components.

    If self already has an existing component of the exact same type, that
    existing component is removed right before adding the provided component
    This is unless the two components are equivalent or the same object,
    in which case the provided component is skipped without posting events.

    If self is in an EntityManager, each component added and removed
    generates a ComponentAdded and ComponentRemoved event, respectively,
    in the manager event queue.

    Args:
      *components: The components to update self with.
    """
    self_components = self._components
    manager = self._manager()
    for component in components:
      component_type = type(component)
      if component_type in self_components:
        other_component = self_components[component_type]
        if other_component is component or other_component == component:
          continue
        if manager is not None:
          manager._component_removed(self, other_component)
      self_components[component_type] = component
      if manager is not None:
        manager._component_added(self, component)

  def remove(self, *components: object) -> None:
    """Remove an arbitrary number of components from self.

    For each provided component, an existing component in self of
    the exact same type is removed. The existing component
    must be equivalent or the same object as the provided component.

    If no such component is found in self, an exception is raised,
    preventing the rest of the components from being removed but
    not affecting the components already removed.

    If self is in an EntityManager, a ComponentRemoved event is
    posted in the manager event queue for each component removed.

    Args:
      *components: The components to be removed.

    Raises:
      ValueError: If one of the components is not found in self.
    """
    self_components = self._components
    manager = self._manager()
    for component in components:
      component_type = type(component)
      try:
        other_component = self_components[component_type]
      except KeyError:
        pass
      else:
        if other_component is component or other_component == component:
          del self_components[component_type]
          if manager is not None:
            manager._component_removed(self, other_component)
          continue
      raise ValueError(component)

  def __getitem__(self, component_type: type[_T], /) -> _T:
    return self._components[component_type]  # type: ignore[return-value]

  def __setitem__(self, component_type: type[_T], component: _T):
    if type(component) is not component_type:
      raise TypeError(component)
    self.update(component)

  def __delitem__(self, component_type: type, /):
    self.remove(self[component_type])

  @overload
  def get(self, component_type: type[_T], /) -> _T | None: ...
  @overload
  def get(self, component_type: type[_T], default: _D, /) -> _T | _D: ...
  def get(self, component_type, default=None, /):
    return self._components.get(component_type, default)

  @overload
  def pop(self, component_type: type[_T], /) -> _T: ...
  @overload
  def pop(self, component_type: type[_T], default: _D, /) -> _T | _D: ...
  def pop(self, component_type, default=_SENTINEL, /):
    try:
      component = self._components[component_type]
    except KeyError:
      if default is _SENTINEL:
        raise
      return default
    self.remove(component)
    return component

  def types(self):
    return self._components.keys()

  def __iter__(self):
    return iter(self._components.values())

  def __reversed__(self):
    return reversed(self._components.values())

  def __len__(self):
    return len(self._components)

  def __contains__(self, obj: object, /):
    return obj in self._components

  def __eq__(self, other: object, /):
    if self is other:
      return True
    if isinstance(other, Entity):
      return self._components == other._components
    return NotImplemented

  def __repr__(self):
    return f'{type(self).__name__}({list(self)})'

  def clear(self) -> None:
    self.remove(*self)

  @staticmethod
  def new_id() -> EntityId:
    """Return a new unique EntityId int.

    The default implementation uses the `uuid` standard
    library, returning the integer value of `uuid4()` which
    is 128 bits long (16 bytes), the UUID standard.
    Subclasses may implement their own generation method.

    This method is used by the Entity's `__init__()` method
    to generate the `id` attribute.

    Returns:
      A unique integer.
    """
    return uuid4().int  # type: ignore[return-value]



"""This module implements the Space class."""

__all__ = ['Space']

from collections import deque
from collections.abc import Callable

from pyriak import EventQueue
from pyriak.entity_manager import EntityManager, QueryResult
from pyriak.state_manager import StateManager
from pyriak.system_manager import SystemManager


class Space:
  """An encapsulation of a standalone program.

  A Space glues together the three managers into
  a single object that represents the entire program.
  It also has a central event queue that controls the flow of the program.

  As this object contains everything, it can be used to access any needed data,
  and is useful to pass as an argument into a function to allow it to
  read or change the program state.

  A Space does not implement anything on its own. Instead, it serves as a
  bundle for the behavior (systems), data (entities and states), and
  communication (events).

  Attributes:
    event_queue: A mutable sequence that holds events to be processed.
    systems: A SystemManager, holds the systems of the space.
    entities: An EntityManager, holds the entities of the space.
    states: A StateManager, holds the states of the space.
  """

  __slots__ = 'event_queue', 'systems', 'entities', 'states', '__weakref__'

  def __init__(
    self, *,
    event_queue: EventQueue | None = None,
    systems: SystemManager | None = None,
    entities: EntityManager | None = None,
    states: StateManager | None = None,
  ):
    """Initialize the Space with the managers and event queue.

    By default (or when None is given as the argument), instantiates
    an empty deque for the event queue, and creates the managers.

    The managers' event_queue attributes are set to the space's
    event queue. The SystemManager's space attribute is set to self.

    Args:
      event_queue: The Space's event queue. Defaults to a creating a new deque.
      systems: The Space's SystemManager. Defaults to creating one.
      entities: The Space's EntityManager. Defaults to creating one.
      states: The Space's StateManager. Defaults to creating one.
    """
    if event_queue is None:
      event_queue = deque()
    self.event_queue = event_queue
    if systems is None:
      systems = SystemManager()
    self.systems = systems
    systems.space = self
    systems.event_queue = event_queue
    if entities is None:
      entities = EntityManager()
    self.entities = entities
    entities.event_queue = event_queue
    if states is None:
      states = StateManager()
    self.states = states
    states.event_queue = event_queue

  def query(
    self, /, *component_types: type, merge: Callable[..., set] = set.intersection
  ) -> QueryResult:
    """Get bulk entity and component data from self's entities.

    Syntactic sugar for self.entities.query().
    For more details and specifics, see documentation of EntityManager.query().

    Args:
      *component_types: The types that are used to generate the set of entities.
      merge: The set merge function used to combine the sets of ids into one.

    Returns:
      A readonly QueryResult object that contains the data and info of the query.

    Raises:
      TypeError: If exactly zero component types were given.

    Example:
      Typical usage of query() method::

        for sprite, position in space.query(Sprite, Position).zip():
          render(sprite, position)
    """
    return self.entities.query(*component_types, merge=merge)

  def process(self, event: object, /) -> bool:
    """Immediately invoke event handlers for an event.

    Syntactic sugar for self.systems.process().
    See SystemManager.process() documentation for more details.

    The difference between process() and post() is that process()
    is synchronous and blocks until the event has been processed,
    while post() is asynchronous, deferring the event to the event queue
    to eventually be processed.
    Since post() only puts events on a queue, it is non-blocking.

    Args:
      event: The event to process.

    Returns:
      True if event processing was stopped by a callback, False otherwise.

    Raises:
      RuntimeError: If the SystemManager's space is None or deleted.
    """
    return self.systems.process(event)

  def post(self, *events: object) -> None:
    """Append an event or events to the end of self's event queue.

    Args:
      *events: The events to be posted to the event queue.
    """
    self.event_queue.extend(events)

  def pump(self, events: int | None = None, /) -> int:
    """Pop and process events from self's event queue.

    For the given number of times, or until the event queue is empty,
    events are popped from the front of the event queue and then processed.
    (The event queue can still get new events while this method is running.)

    The number of times defaults to None (infinite), meaning it will
    only stop when the event queue is empty.

    pump() is safe to be called recursively/nested.

    The return value is the actual number of events processed, which is
    less than or equal to the number passed in.
    It is 0 if and only if the event queue was already empty when pump()
    was called and there are no more events left to process.

    Args:
      events: The max number of events to process. Defaults to None (infinite).

    Returns:
      The number of events actually popped from the event queue and processed.
    """
    process_event = self.process
    queue = self.event_queue
    pop = queue.popleft if isinstance(queue, deque) else lambda: queue.pop(0)
    i = -1
    if events is None:
      while queue:
        i += 1
        process_event(pop())
    else:
      for i in range(events):
        if not queue:
          return i
        process_event(pop())
    return i + 1



"""This module contains the implementation for binding event handlers."""

__all__ = ['bind', 'Binding']

from collections.abc import Callable, Hashable, Iterable
from functools import update_wrapper
from typing import (
  TYPE_CHECKING,
  Any,
  Generic,
  TypeAlias,
  TypeVar,
  overload,
)

from pyriak import _SENTINEL
from pyriak.eventkey import key_functions


if TYPE_CHECKING:
  from pyriak.space import Space


_T = TypeVar('_T')
_R_co = TypeVar('_R_co', covariant=True)

_Callback: TypeAlias = Callable[['Space', _T], _R_co]


_empty_frozenset: frozenset[object] = frozenset()


class Binding(Generic[_T, _R_co]):
  """A Binding wraps the event handler callback with handler info.

  bind() returns a Binding. When a system is added to a SystemManager,
  it searches the system for attributes of type Binding.
  An event handler is then created for each binding on the system.

  Binding forwards calls to the internal callback.
  Binding supports descriptor access by redirecting it to the internal
  callback. This is to allow instance methods to be properly invoked as a
  bound method (for the self argument), if the callback was a function
  in a class.

  Attributes:
    _callback_: The event handler callback wrapped by the Binding.
    _event_type_: The event type bound to the handler.
    _priority_: The priority of the handler for this event type.
    _keys_: The frozenset of event keys that the handler be triggered by.
      Often empty, or only containing one key.
  """

  __wrapped__: _Callback[_T, _R_co]

  def __init__(
    self,
    callback: _Callback[_T, _R_co],
    event_type: type[_T],
    priority: Any,
    keys: frozenset[Hashable],
  ):
    update_wrapper(self, callback)
    self._event_type_ = event_type
    self._priority_ = priority
    self._keys_ = keys

  @property
  def _callback_(self) -> _Callback[_T, _R_co]:
    return self.__wrapped__

  def __call__(self, space: 'Space', event: _T, /) -> _R_co:
    return self._callback_(space, event)

  def __get__(self, obj, objtype=None):
    callback = self._callback_
    try:
      descr_get = type(callback).__get__  # type: ignore[attr-defined]
    except AttributeError:
      return callback
    return descr_get(callback, obj, objtype)


@overload
def bind(
  event_type: type[_T], priority: Any, /
) -> Callable[[_Callback[_T, _R_co]], Binding[_T, _R_co]]: ...
@overload
def bind(
  event_type: type[_T], priority: Any, /, *, key: Hashable
) -> Callable[[_Callback[_T, _R_co]], Binding[_T, _R_co]]: ...
@overload
def bind(
  event_type: type[_T], priority: Any, /, *, keys: Iterable[Hashable]
) -> Callable[[_Callback[_T, _R_co]], Binding[_T, _R_co]]: ...
def bind(event_type, priority, /, *, key=_SENTINEL, keys=_SENTINEL):
  """Bind a callback to an event type.

  To use, define a function that takes two arguments, space and event.
  Then, decorate it with a call to bind(), passing in the necessary info.

  This creates a binding. When the system is added to the space's SystemManager,
  an event handler will be created for this binding, which means that when an event
  of the correct type is processed by the SystemManager, the callback is invoked.
  This does not include subclasses of the event type. The types must be exact.

  The priority determines the order in which callbacks are invoked if there
  are multiple systems or event handlers for that event.

  The optional key or keys further narrow which events are handled.
  The event must give at least one key that matches the binding,
  if the binding has any keys.
  This is only valid for event types with key functions.

  bind() should only be used on attributes directly on the system.
  bind() cannot be used multiple times on the same object.

  bind() works on any callable, but the signature should be correct.
  It can be manually invoked with two calls instead of using as a decorator,
  if necessary.
  The most common place where bind() is used is on a module top-level function,
  where the module object is the system.

  Args:
    event_type: The type of events that the handler will be triggered by.
    priority: The object the handler will be sorted by for invocation.
    key: Defaults to no key. The key that events must have for the handler.
    keys: Defaults to no keys. The keys that events must have any of.

  Returns:
    A Binding instance that allows SystemManager to recognize bindings.

  Raises:
    TypeError: If the argument types or function call signature are incorrect.
      If `event_type` is not a type object and hashable.
      If key(s) were provided but the event type doesn't have a key function.
      If both `key` and `keys` keyword arguments are passed in.
      If any of the keys provided are not hashable.
      In the decorator, if the object is already a binding.

  Example:
    Typical usage of bind() decorator::

      @bind(UpdateGame, 500)
      def update_physics(space: Space, event: UpdateGame):
        ...
  """
  if not isinstance(event_type, type):
    raise TypeError(f'{event_type!r} is not a type')
  try:
    hash(event_type)
  except TypeError:
    raise TypeError(f'{event_type!r} is not hashable') from None
  if key is not _SENTINEL:
    if keys is not _SENTINEL:
      raise TypeError("bind() cannot be passed both 'key' and 'keys' kwargs")
    keys = frozenset([key])
  else:
    keys = frozenset(keys) if keys is not _SENTINEL else _empty_frozenset
  if keys and event_type not in key_functions:
    raise TypeError(
      f'bind(): keys were provided but no key function exists for {event_type!r}'
    )
  def decorator(callback, /):
    if isinstance(callback, Binding):
      raise TypeError('cannot bind same object multiple times')
    return Binding(callback, event_type, priority, keys)
  return decorator



"""This module implements the functionality responsible for keyed event types.

Along with the event type, a bound event handler can further narrow down
what events it receives with an event key or keys.
An event's key, determined by the key function of the event type,
must match at least one of the bound keys.

A key function takes one argument, the event,
and returns either one key or an iterator of any number of keys.
Keys must be hashable.

Binding keys only available for event types with key functions,
but are not required. If no keys are bound, the event handler
acts as normal, ignoring event keys.

An event type either has or doesn't have a key function, and
this is global across the entire python process.
A key function cannot be removed once set.

To set a key function on an event type, either use the
key_functions variable as a mapping to do it manually,
or use the set_key() function/decorator on the class (preferred).

The operator module has useful functions for
creating key functions, e.g. attrgetter.
"""

__all__ = ['key_functions', 'set_key', 'KeyFunction']

from collections.abc import Callable, Hashable, Iterator, Mapping
from typing import Any, TypeAlias, TypeVar, overload
from weakref import WeakKeyDictionary


_T = TypeVar('_T')
_D = TypeVar('_D')

del TypeVar


KeyFunction: TypeAlias = Callable[[_T], Hashable | Iterator[Hashable]]
"""Type alias for a valid key function for a specified event type.

If the return type is an iterator (especially generator), the event keys
are what the iterator yields. Otherwise, the event key is the return value.
All keys must be hashable.
"""


class EventKeyFunctions:
  """A dictionary of event types to key functions.

  Once a key function is set for an event type, it cannot be
  deleted or replaced.
  Currently, there is only one key function mapping per process,
  accessible as pyriak.key_functions.

  Internally, EventKeyFunctions is implemented using a WeakKeyDictionary,
  meaning that event types are not kept alive just because they
  have a key function. Since type objects usually don't get deleted,
  this feature may rarely be used.
  Note: In CPython, type objects have reference cycles, so only the garbage
  collector can delete them, not reference counting.
  """

  __slots__ = ('_data',)

  def __init__(
    self, dict: Mapping[type, KeyFunction[Any]] | None = None
  ):
    self._data: WeakKeyDictionary[type, KeyFunction[Any]]
    self._data = WeakKeyDictionary()
    if dict is not None:
      self.update(dict)

  def __getitem__(self, event_type: type[_T]) -> KeyFunction[_T]:
    return self._data[event_type]

  def __setitem__(self, event_type: type[_T], key: KeyFunction[_T]):
    """Set a key function for an event type.

    Raises:
      ValueError: If the given event type already has a key function set.
    """
    data = self._data
    if event_type in data:
      raise ValueError(f'cannot reassign key function for event type {event_type!r}')
    data[event_type] = key

  @overload
  def get(self, event_type: type[_T]) -> KeyFunction[_T] | None: ...
  @overload
  def get(self, event_type: type[_T], default: _D) -> KeyFunction[_T] | _D: ...
  def get(self, event_type, default=None):
    return self._data.get(event_type, default)

  def setdefault(
    self, event_type: type[_T], default: KeyFunction[_T]
  ) -> KeyFunction[_T]:
    data = self._data
    if event_type in data:
      return data[event_type]
    data[event_type] = default
    return default

  def update(self, other: Mapping[type, KeyFunction[Any]]) -> None:
    for event_type, key in dict(other).items():
      self[event_type] = key

  def keys(self):
    return self._data.keys()

  def values(self):
    return self._data.values()

  def items(self):
    return self._data.items()

  def __len__(self):
    return len(self._data)

  def __contains__(self, event_type: type):
    return event_type in self._data

  def __iter__(self):
    return iter(self._data)

  def __or__(self, other: Mapping[type, KeyFunction[Any]]):
    return self._data | other

  def __ror__(self, other: Mapping[type, KeyFunction[Any]]):
    return other | self._data

  def __ior__(self, other: Mapping[type, KeyFunction[Any]]):
    self.update(other)
    return self

  def copy(self):
    return self._data.copy()

  def keyrefs(self):
    return self._data.keyrefs()


key_functions = EventKeyFunctions()


def set_key(key: KeyFunction[_T], /) -> Callable[[type[_T]], type[_T]]:
  """Assign a key function to an event type.

  Use as a decorator to assign a key function to an event class definition.

  Args:
    key: The key function to be assigned.

  Returns:
    A decorator that sets the key function on its argument
    and then returns the argument.
    This decorator raises ValueError if the type already has a key function.
  """
  def decorator(cls: type[_T]) -> type[_T]:
    key_functions[cls] = key
    return cls
  return decorator



# ----------------------------------------------------------------------------
# pyriak
# Copyright (c) 2024 aatle

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ----------------------------------------------------------------------------

"""Pyriak is a framework for ECS architecture.

This implementation also implements event-driven architecture.
The package is written in pure python and has no dependencies.

This module is the top-level module of the package and contains
most of the things necessary to use this package.
"""

__all__ = [
  'bind',
  'Space',
  'Entity',
  'EntityId',
  'System',
  'EventQueue',
  'key_functions',
  'set_key',
]

from collections.abc import Hashable as _Hashable, MutableSequence as _MutableSequence
from enum import Enum as _Enum
from typing import Any as _Any, TypeAlias as _TypeAlias
from weakref import ref as _weakref


System: _TypeAlias = _Hashable


EventQueue: _TypeAlias = _MutableSequence[object]


dead_weakref: _weakref[_Any] = _weakref(set())


class _Sentinel(_Enum):
  SENTINEL = 1

_SENTINEL: _Sentinel = _Sentinel.SENTINEL


# circular imports
from pyriak.bind import bind  # noqa: E402
from pyriak.entity import Entity, EntityId  # noqa: E402
from pyriak.entity_manager import QueryResult as QueryResult  # noqa: E402
from pyriak.eventkey import key_functions, set_key  # noqa: E402
from pyriak.space import Space  # noqa: E402


# cleanup namespace
del _Hashable, _MutableSequence, _Enum, _weakref



"""This module contains built-in events.

The managers automatically generate these events when
something is added or removed from them, if they have an event queue.
Entities added to an EntityManager also generate events for components.

Most of these events have key functions to help narrow down which events
to receive, since they are generated in large volumes and very generally.
"""

__all__ = [
  'ComponentAdded',
  'ComponentRemoved',
  'EntityAdded',
  'EntityRemoved',
  'SystemAdded',
  'SystemRemoved',
  'EventHandlerAdded',
  'EventHandlerRemoved',
  'StateAdded',
  'StateRemoved',
]

from collections.abc import Hashable
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from pyriak.eventkey import set_key as _set_key


if TYPE_CHECKING:
  from pyriak import System
  from pyriak.bind import Binding, _Callback
  from pyriak.entity import Entity
  from pyriak.system_manager import _EventHandler


_T = TypeVar('_T')


class EntityAdded:
  """An event for when an Entity is added to the EntityManager.

  Attributes:
    entity: The entity added to the manager.
  """

  def __init__(self, entity: 'Entity'):
    self.entity = entity


class EntityRemoved:
  """An event for when an Entity is removed from the EntityManager.

  Attributes:
    entity: The entity removed from the manager.
  """

  def __init__(self, entity: 'Entity'):
    self.entity = entity


def _component_type_key(event: 'ComponentAdded | ComponentRemoved') -> type:
  return type(event.component)


@_set_key(_component_type_key)
class ComponentAdded(Generic[_T]):
  """An event for when a component is added to the EntityManager.

  Either, a component is added to an entity that is in the manager,
  or an entity with existing components is added to the manager.

  The event key is the type of the component.

  Attributes:
    entity: The entity that the component is in.
    component: The component added.
  """

  def __init__(self, entity: 'Entity', component: _T):
    self.entity = entity
    self.component = component


@_set_key(_component_type_key)
class ComponentRemoved(Generic[_T]):
  """An event for when a component is removed from the EntityManager.

  Either, a component is removed from an entity that is in the manager,
  or an entity with components is removed from the manager.
  For the latter case, the component may still be on the entity.

  The event key is the type of the component.

  Attributes:
    entity: The entity that the component was in.
    component: The component removed.
  """

  def __init__(self, entity: 'Entity', component: _T):
    self.entity = entity
    self.component = component


def _system_key(event: 'SystemAdded | SystemRemoved') -> 'System':
  return event.system


@_set_key(_system_key)
class SystemAdded:
  """An event for when a system is added to the SystemManager.

  The event key is the system itself.

  Attributes:
    system: The system added to the manager.
  """

  def __init__(self, system: 'System'):
    self.system = system


@_set_key(_system_key)
class SystemRemoved:
  """An event for when a system is removed from the SystemManager.

  The event key is the system itself.

  Attributes:
    system: The system removed from the manager.
  """

  def __init__(self, system: 'System'):
    self.system = system


def _state_type_key(event: 'StateAdded | StateRemoved') -> type:
  return type(event.state)


@_set_key(_state_type_key)
class StateAdded(Generic[_T]):
  """An event for when a state is added to the StateManager.

  The event key is the type of the state.

  Attributes:
    state: The state added to the manager.
  """

  def __init__(self, state: _T):
    self.state = state


@_set_key(_state_type_key)
class StateRemoved(Generic[_T]):
  """An event for when a state is removed from the StateManager.

  The event key is the type of the state.

  Attributes:
    state: The state removed from the manager.
  """

  def __init__(self, state: _T):
    self.state = state


def _handler_key(event: 'EventHandlerAdded | EventHandlerRemoved') -> type:
  return event.event_type


class _EventHandlerEvent(Generic[_T]):
  def __init__(
    self, _binding: 'Binding[_T, Any]', _handler: '_EventHandler[_T]'
  ):
    self._binding = _binding
    self._handler = _handler

  @property
  def system(self) -> 'System':
    return self._handler.system

  @property
  def callback(self) -> '_Callback[_T, Any]':
    return self._handler.callback

  @property
  def name(self) -> str:
    return self._handler.name

  @property
  def priority(self) -> Any:
    return self._binding._priority_

  @property
  def event_type(self) -> type[_T]:
    return self._binding._event_type_

  @property
  def keys(self) -> frozenset[Hashable]:
    return self._binding._keys_

  @property
  def key(self) -> Hashable:
    """The single key of the event handler, if applicable.

    Often, a key event handler only binds a single key.
    Raises ValueError if there is not exactly 1 key in keys.
    """
    [key] = self.keys
    return key


@_set_key(_handler_key)
class EventHandlerAdded(_EventHandlerEvent[_T]):
  """An event for when a single event handler is added.

  When a system is added to the SystemManager, it may have bindings.
  For each binding, an event handler is created on the manager.

  The event key is the event type of the handler.

  Attributes:
    system: The system of the event handler.
    callback: The callback of the event handler.
    name: The attribute name of the binding on the system.
    priority: The priority of the event handler.
    event_type: The event type of the event handler.
    keys: The keys of the event handler. May be empty.
  """


@_set_key(_handler_key)
class EventHandlerRemoved(_EventHandlerEvent[_T]):
  """An event for when a single event handler is removed.

  A system removed from the SystemManager may own event
  handlers that need to be removed.

  The event key is the event type of the handler.

  Attributes:
    system: The system of the event handler.
    callback: The callback of the event handler.
    name: The attribute name of the binding on the system.
    priority: The priority of the event handler.
    event_type: The event type of the event handler.
    keys: The keys of the event handler. May be empty.
  """



"""This module implements the EntityManager and its helper classes."""

__all__ = ['EntityManager', 'QueryResult']

from collections.abc import Collection, Iterable, Iterator, KeysView, Set as AbstractSet
from typing import Any, Callable, TypeVar, overload
from weakref import ref as weakref

from pyriak import _SENTINEL, EventQueue, dead_weakref
from pyriak.entity import Entity, EntityId
from pyriak.events import ComponentAdded, ComponentRemoved, EntityAdded, EntityRemoved


_T = TypeVar('_T')


class QueryResult:
  """An object holding the results of the query, with methods to access it.

  Data requested from the EntityManager is exposed in a couple different
  formats, using the QueryResult object.
  It also contains the parameters for the query. The types may be used
  as a default for certain methods.

  Since the data is computed only once, a QueryResult should generally not be
  stored as it may become stale. There is little reason to keep it alive
  outside of the query call expression that created it.
  """

  __slots__ = '_entities', '_types', '_merge'

  def __init__(
    self, _entities: dict[EntityId, Entity],
    _types: tuple[type, ...],
    _merge: Callable[..., set], /
  ):
    self._entities = _entities
    self._types = _types
    self._merge = _merge

  @property
  def ids(self) -> AbstractSet[EntityId]:
    """The ids of the entities of the query."""
    return self._entities.keys()

  @property
  def entities(self) -> Collection[Entity]:
    """The entities of the query."""
    return self._entities.values()

  @property
  def types(self) -> tuple[type, ...]:
    """The component types passed to the query call, in order."""
    return self._types

  @property
  def merge(self) -> Callable[..., set]:
    """The merge function passed to the query call.

    The default query merge function is set.intersection.
    """
    return self._merge

  def __call__(self, component_type: type[_T], /) -> Iterator[_T]:
    return (entity[component_type] for entity in self.entities)

  def zip(self, *component_types: type) -> Iterator[tuple[Any, ...]]:
    """Return an iterator of tuples of components.

    For each entity, a tuple is yielded with components corresponding to the
    component types provided. If no component types are given, it defaults
    to the component types passed in for the query call, self.types.

    If the entity doesn't have one of the component types when the entity is
    reached in the iterator, a KeyError is raised from the entity.

    Args:
      *component_types: The types to access from the query's entities.

    Returns:
      An iterator yielding same-length tuples of components.

    Example:
      Typical usage of zip() method::

        for sprite, position in space.query(Sprite, Position).zip():
          render(sprite, position)
    """
    if not component_types:
      component_types = self.types
    return (
      tuple([ent[comp_type] for comp_type in component_types]) for ent in self.entities
    )

  def zip_entity(self, *component_types: type) -> Iterator[tuple[Any, ...]]:
    """Return an iterator of tuples of the entity and its components.

    This method is the same as zip() except that each tuple starts with the
    entity, which is followed by the components.
    The component types default to self.types, the types in the query call,
    if none are provided.

    Args:
      *component_types: The types to access from the query's entities.

    Returns:
      An iterator yielding same-length tuples of the entity and its components.

    Example:
      Typical usage of zip_entity() method::

        for entity, lifetime in space.query(Lifetime).zip_entity():
          if lifetime.timer < 0:
            space.post(KillEntity(entity))
    """
    if not component_types:
      component_types = self.types
    return (
      (ent, *[ent[comp_type] for comp_type in component_types]) for ent in self.entities
    )

  __hash__ = None  # type: ignore[assignment]


class _Components:
  """The components attribute of an EntityManager, exposes more functionality.

  The EntityManager not only manages the entities it holds, but also partially
  manages the components in those entities.
  The components attribute of the EntityManager exposes additional operations for
  the components in the entities, while the entity operations are on the manager.

  Instances are not meant to be independent; they serve as an extension
  to the EntityManager in a syntactically intuitive way, through dot access.
  As such, they are only a proxy for the manager's data.

  If the EntityManager has been deleted and garbage collected, all methods
  raise TypeError. A _Components instance only weakly references its manager.

  Attributes:
    _manager: A weakref to the manager (narrower type annotation to avoid errors).
  """

  __slots__ = ('_manager',)

  _manager: Callable[[], 'EntityManager']

  def __init__(self, manager: 'EntityManager', /):
    self._manager = weakref(manager)  # type: ignore[assignment]

  def __call__(self, component_type: type[_T], /) -> Iterator[_T]:
    """Return an iterator of all components of a type in the manager.

    For each entity that has it, its component is yielded.
    The order of entities/components is not defined.
    The iterator may be empty.

    Since it is an iterator, there are some limitations with it:
    - Raises KeyError if one of the entities is no longer
      in the manager when it is reached in iteration.
    - Raises RuntimeError if the number of entities with
      the component type changes during iteration.
    If necessary, these can be avoided by converting to tuple.

    Args:
      component_type: The type of components to get.

    Returns:
      An iterator of components of that type, from the manager's entities.
    """
    manager = self._manager()
    entities = manager._entities
    return (
      entities[entity_id][component_type]
      for entity_id in manager._type_cache.get(component_type, ())
    )

  def types(self) -> KeysView[type]:
    return self._manager()._type_cache.keys()

  def __iter__(self) -> Iterator[object]:
    for entity in self._manager():
      yield from entity

  def __reversed__(self) -> Iterator[object]:
    for entity in reversed(self._manager()):
      yield from reversed(entity)

  def __len__(self):
    return sum([len(entity) for entity in self._manager()])

  def __contains__(self, obj: object, /):
    return obj in self._manager()._type_cache


class EntityManager:
  """A manager and container of entities.

  The EntityManager stores the entities of a space and provides ways to
  access them and their components.
  Almost all entities instantiated get added to the EntityManager
  so they are accessible from the rest of the space.

  An entity may only be stored in at most one EntityManager at once,
  because entities weakly reference their managers to update caches.

  The manager also has a .components attribute which extends the manager by
  exposing additional (readonly) operations for the components within the entities.

  The EntityManager is owned directly by the Space.
  There is usually only one, accessible via the Space's 'entities' attribute.

  Attributes:
    event_queue: The optional event queue that the EntityManager may post to.
      This is usually assigned the space's event queue.
  """

  __slots__ = '_entities', '_type_cache', 'components', 'event_queue', '__weakref__'

  def __init__(
    self, entities: Iterable[Entity] = (), /, event_queue: EventQueue | None = None
  ):
    """Initialize the EntityManager with the given entities and event queue.

    By default, the EntityManager is initialized with no entities and
    event queue as None.

    Also initializes the components attribute of the manager.

    Args:
      entities: The iterable of initial entities. Defaults to no entities.
      event_queue: The event queue to post to. Defaults to None.
    """
    self.event_queue = event_queue
    self.components = _Components(self)
    self._entities: dict[EntityId, Entity] = {}
    self._type_cache: dict[type, set[EntityId]] = {}
    self.add(*entities)

  def add(self, *entities: Entity) -> None:
    """Add an arbitrary number of entities to self.

    Each entity is stored by its unique id.

    If self already has the entity being added, an exception is raised.

    If self's event queue is not None, each entity added generates
    an EntityAdded event and then a ComponentAdded event for each
    component in the entity.

    An entity may only added to at most one manager.

    Args:
      *entities: The entities to be added.

    Raises:
      ValueError: If self already has one of the entities.
      RuntimeError: If one of the entities is already added to another manager.
    """
    component_types = self._type_cache
    self_entities = self._entities
    self_weakref: weakref[EntityManager] = weakref(self)
    event_queue = self.event_queue
    for entity in entities:
      entity_id = entity.id
      if entity_id in self_entities:
        raise ValueError(entity)
      if entity._manager() is not None:
        raise RuntimeError(f'{entity!r} already added to another manager')
      self_entities[entity_id] = entity
      entity._manager = self_weakref
      for component_type in entity.types():
        try:
          component_types[component_type].add(entity_id)
        except KeyError:
          component_types[component_type] = {entity_id}
      if event_queue is not None:
        event_queue.extend([
          EntityAdded(entity),
          *[ComponentAdded(entity, component) for component in entity]
        ])

  def update(self, *entities: Entity) -> None:
    """Update self with an arbitrary number of entities.

    Same as add(), except that entities that are already in self
    get skipped instead of raising an error.

    Args:
      *entities: The entities to update self with.

    Raises:
      RuntimeError: If one of the entities is in another manager.
    """
    for entity in entities:
      if entity not in self:
        self.add(entity)

  def create(self, *components: object) -> Entity:
    """Create and return a new Entity with components, added to self.

    The entity is instantiated with the given components.
    Then, it is added to self and returned.
    The adding generates EntityAdded and ComponentAdded events.

    Args:
      *components: The components of the new entity.

    Returns:
      The entity created.
    """
    entity: Entity = Entity(components)
    self.add(entity)
    return entity

  def remove(self, *entities: Entity) -> None:
    """Remove an arbitrary number of entities from self.

    If the entity being removed is not in self, a ValueError
    is raised, preventing the rest of the entities from being removed,
    but not affecting the entities already removed.

    If self's event queue is not None, for each entity, a ComponentRemoved
    event is posted for each component followed by an EntityRemoved event.

    Args:
      *entities: The entities to be removed.

    Raises:
      ValueError: If one of the entities is not in self.
    """
    type_cache = self._type_cache
    event_queue = self.event_queue
    for entity in entities:
      entity_id = entity.id
      try:
        del self._entities[entity_id]
      except KeyError:
        raise ValueError(entity) from None
      entity._manager = dead_weakref
      for component_type in entity.types():
        component_type_entities = type_cache[component_type]
        component_type_entities.remove(entity_id)
        if not component_type_entities:
          del type_cache[component_type]
      if event_queue is not None:
        event_queue.extend([
          *[ComponentRemoved(entity, component) for component in entity],
          EntityRemoved(entity)
        ])

  def query(
    self, /, *component_types: type, merge: Callable[..., set] = set.intersection
  ) -> QueryResult:
    """Request specific data from the manager.

    Given an arbitrary non-zero number of types, return bulk data about
    the set of entities corresponding to these types, as a QueryResult.

    Each component type has a set of entities/ids that have it.
    The merge function combines these sets to create the final set of
    entities/ids that are used.

    By default, the merge function is set.intersection, which means the
    result is the entities that have all of the component types.

    The merge function must take a number of sets as arguments, where the
    number is how many component types are passed in, and return a set.
    Preferably, the merge function can take an arbitrary non-zero number of sets.
    The most common merge functions are unbound set methods.

    Args:
      *component_types: The types that are used to generate the set of entities.
      merge: The set merge function used to combine the sets of ids into one.

    Returns:
      A readonly QueryResult object that contains the data and info of the query.

    Raises:
      TypeError: If exactly zero component types were given.
    """
    if not component_types:
      raise TypeError('expected at least one component type')
    self_entities = self._entities
    type_cache = self._type_cache
    return QueryResult(
      {
        id: self_entities[id]
        for id in merge(*[
          type_cache[typ]  # noqa: SIM401
          if typ in type_cache else set()
          for typ in component_types
        ])
      },
      component_types,
      merge
    )

  def __call__(self, component_type: type, /) -> Iterator[Entity]:
    """Return an iterator of all entities with component type in the manager.

    For each entity that has a component of the type given, it is yielded.
    The order of entities is not defined. The iterator may be empty.

    Since it is an iterator, there are some limitations with it:
    - Raises KeyError if one of the entities is no longer
      in the manager when it is reached in iteration.
    - Raises RuntimeError if the number of entities with
      the component type changes during iteration.
    If necessary, these can be avoided by converting to tuple.

    Args:
      component_type: The component type to find entities with.

    Returns:
      An iterator of entities that have component_type.
    """
    entities = self._entities
    return (
      entities[entity_id] for entity_id in self._type_cache.get(component_type, ())
    )

  def __getitem__(self, entity_id: EntityId, /) -> Entity:
    return self._entities[entity_id]

  def __delitem__(self, entity_id: EntityId, /):
    self.remove(self._entities[entity_id])

  @overload
  def get(self, entity_id: EntityId, /) -> Entity | None: ...
  @overload
  def get(self, entity_id: EntityId, default: _T, /) -> Entity | _T: ...
  def get(self, entity_id, default=None, /):
    return self._entities.get(entity_id, default)

  @overload
  def pop(self, entity_id: EntityId, /) -> Entity: ...
  @overload
  def pop(self, entity_id: EntityId, default: _T, /) -> Entity | _T: ...
  def pop(self, entity_id, default=_SENTINEL, /):
    try:
      entity = self._entities[entity_id]
    except KeyError:
      if default is _SENTINEL:
        raise
      return default
    self.remove(entity)
    return entity

  @overload
  def ids(self, /) -> KeysView[EntityId]: ...
  @overload
  def ids(self, component_type: type, /) -> set[EntityId]: ...
  def ids(self, component_type=None, /):
    """Return a set of all entity ids that contain the given component type.

    If no component type is given or is None, then return a readonly set-like
    view of all entity ids in self instead (keys view).

    Args:
      component_type: The type that entities must have. Defaults to None.

    Returns:
      Either a set-like keys view of ids or a set of ids.
    """
    if component_type is None:
      return self._entities.keys()
    return set(self._type_cache.get(component_type, ()))

  def __iter__(self):
    return iter(self._entities.values())

  def __reversed__(self):
    return reversed(self._entities.values())

  def __len__(self):
    return len(self._entities)

  def __contains__(self, obj: object, /):
    if isinstance(obj, Entity):
      return obj.id in self._entities
    return obj in self._entities

  def clear(self) -> None:
    self.remove(*self)

  def _component_added(self, entity: Entity, component: object, /) -> None:
    try:
      self._type_cache[type(component)].add(entity.id)
    except KeyError:
      self._type_cache[type(component)] = {entity.id}
    event_queue = self.event_queue
    if event_queue is not None:
      event_queue.append(ComponentAdded(entity, component))

  def _component_removed(self, entity: Entity, component: object, /) -> None:
    entity_ids = self._type_cache[type(component)]
    entity_ids.remove(entity.id)
    if not entity_ids:
      del self._type_cache[type(component)]
    event_queue = self.event_queue
    if event_queue is not None:
      event_queue.append(ComponentRemoved(entity, component))



"""This module implements the SystemManager class."""

__all__ = ['SystemManager']

from collections.abc import Hashable, Iterable, Iterator
from inspect import getattr_static
from types import ModuleType
from typing import TYPE_CHECKING, Any, Generic, NamedTuple, TypeVar
from weakref import ref as weakref

from pyriak import EventQueue, System, dead_weakref
from pyriak.bind import Binding, _Callback
from pyriak.eventkey import key_functions
from pyriak.events import (
  EventHandlerAdded,
  EventHandlerRemoved,
  SystemAdded,
  SystemRemoved,
)


if TYPE_CHECKING:
  from pyriak.space import Space


_T = TypeVar('_T')


class _EventHandler(NamedTuple, Generic[_T]):
  """Internal object that holds info for a single SystemManager event handler.

  A function on a system can be bound to event types.
  This creates a 'binding' on the system.
  There may be multiple event handlers per binding, each bound to
  different event types and having separate priority and keys.
  The function becomes the event handler callback for each of those handlers.
  Most commonly, a binding only has one event handler.
  Conceptually, an event handler consists of its event type, keys, priority,
  callback, system, and name, and is in a manager listening for events.
  The system, name, and callback are all shared
  in a single binding.
  However, this internal object only stores some of that data,
  as other data is implied by its location in data structures.

  This object is used for invoking, sorting, and storing the event handler.
  It is implemented as a NamedTuple, with equality and hash based on only
  name and system.

  Attributes:
    system: The system the event handler belongs to.
    callback: The event handler callback to be invoked.
    name: The attribute name of the binding on the system.
    priority: The priority of the event handler given in bind().
  """

  system: System
  callback: _Callback[_T, Any]
  name: str
  priority: Any

  def __call__(self, space: 'Space', event: _T, /) -> Any:
    return self.callback(space, event)

  def __eq__(self, other: object):
    if self is other:
      return True
    if isinstance(other, _EventHandler):
      return self.name == other.name and self.system == other.system
    return NotImplemented

  def __hash__(self):
    return hash((self.name,self.system))

del NamedTuple


class SystemManager:
  """A manager and container of systems.

  The SystemManager stores the systems of a space and is in charge of invoking
  their event handlers.
  Event processing and other system callbacks require a reference to the space,
  so the SystemManager holds a weak reference to one.

  A system is any hashable object, but it must have certain things to make it
  useful. Defining _added_ and _removed_ functions or attributes on the system
  executes code when a system is added or removed.
  Use bind() to create event handlers that listen for certain event types.

  Attributes:
    event_queue: The optional event queue that the SystemManager may post to.
      This is usually assigned the space's event queue.
  """

  __slots__ = (
    '_space', '_systems', '_handlers', '_key_handlers', 'event_queue', '__weakref__'
  )

  def __init__(
    self,
    systems: Iterable[System] = (),
    /,
    space: 'Space | None' = None,
    event_queue: EventQueue | None = None
  ):
    """Initialize the SystemManager with systems, space, and event queue.

    By default, the SystemManager is initialized with no systems, event queue
    as None, and space as None.

    Args:
      systems: The iterable of initial systems. Defaults to no systems.
      space: The space to use in system callbacks.
      event_queue: The event queue to post to. Defaults to None.
    """
    self.space = space
    self.event_queue = event_queue
    # values aren't used: dict is for insertion order
    self._systems: dict[System, None] = {}
    self._handlers: dict[type, list[_EventHandler[Any]]] = {}
    self._key_handlers: dict[type, dict[Hashable, list[_EventHandler[Any]]]] = {}
    self.add(*systems)

  def process(self, event: object, /) -> bool:
    """Invoke system event handlers for an event.

    Event handlers listen for events of a specific type.
    Optionally, they can require event keys.
    Each event handler callback is called in order of priority.

    The SystemManager's space attribute must not be None.

    If the event type has no event handlers, nothing happens.

    If a callback returns a truthy value, the rest of the callbacks
    are skipped and True is returned.
    If no callback returns a truthy value, False is returned.

    Args:
      event: The event to process.

    Returns:
      True if event processing was stopped by a callback, False otherwise.

    Raises:
      RuntimeError: If self's space is None or deleted.
    """
    space = self.space
    if space is None:
      raise RuntimeError('cannot process event, space is None or deleted')
    for handler in self._get_handlers(event):  # noqa: SIM110
      if handler.callback(space, event):
        return True
    return False

  def add(self, *systems: System) -> None:
    """Add an arbitrary number of systems and their event handlers to self.

    The systems are added one at a time.
    Any hashable object is a valid system.

    If the system is already in self, an exception is raised.

    If self's event queue is not None, a SystemAdded event is generated,
    followed by EventHandlerAdded events for its event handlers.

    If self's space is not None and the system has an _added_ attribute,
    the attribute is called with the space as the only argument.
    This callback can be used to initialize any necessary things for the system.

    Args:
      *systems: The systems to be added.

    Raises:
      ValueError: If self already has one of the systems.
    """
    self_systems = self._systems
    bind = self._bind
    for system in systems:
      if system in self_systems:
        raise ValueError(f'system manager already has system {system}')
      self_systems[system] = None
      events = bind(system)
      space = self.space
      event_queue = self.event_queue
      if event_queue is not None:
        event_queue.extend([SystemAdded(system), *events])
      if space is not None:
        try:
          added = system._added_  # type: ignore[attr-defined]
        except AttributeError:
          continue
        added(space)

  def update(self, *systems: System) -> None:
    """Update self with an arbitrary number of systems.

    Same as add(), except that systems that are already in self
    get skipped instead of raising an error.

    Args:
      *systems: The systems to update self with.
    """
    for system in systems:
      if system not in self._systems:
        self.add(system)

  def remove(self, *systems: System) -> None:
    """Remove an arbitrary number of systems and their event handlers from self.

    The systems are removed one at a time.

    If the system is not in self, an exception is raised, preventing the rest
    of the systems from being removed.

    If self's event queue is not None, for each system, an EventHandlerRemoved
    event is posted for each of its event handlers followed by a SystemRemoved event.

    If self's space is not None and the system has a _removed_ attribute,
    the attribute is called with the space as the only argument.
    This callback can be used to remove anything tied to the system.

    Args:
      *systems: The systems to be removed.

    Raises:
      ValueError: If one of the systems is not in self.
    """
    self_systems = self._systems
    unbind = self._unbind
    for system in systems:
      try:
        del self_systems[system]
      except KeyError:
        raise ValueError(system) from None
      events = unbind(system)
      space = self.space
      event_queue = self.event_queue
      if event_queue is not None:
        event_queue.extend([*events, SystemRemoved(system)])
      if space is not None:
        try:
          removed = system._removed_  # type: ignore[attr-defined]
        except AttributeError:
          continue
        removed(space)

  def discard(self, *systems: System) -> None:
    """Remove systems, skipping any not in self.

    This method is the same as remove(), with one difference:
    it does not raise an exception when a system is missing from self.
    Instead, the system is skipped.

    See documentation of remove() for more info.

    Args:
      *systems: The systems to be removed if in self.
    """
    self_systems = self._systems
    for system in systems:
      if system in self_systems:
        self.remove(system)

  def __iter__(self):
    return iter(self._systems)

  def __reversed__(self):
    return reversed(self._systems)

  def __len__(self):
    return len(self._systems)

  def __contains__(self, obj: object, /):
    return obj in self._systems

  def clear(self) -> None:
    """Remove all systems and event handlers from self.

    Does not trigger any events or invoke _removed_ callbacks.
    Does not affect self's space or event_queue.
    """
    self._handlers.clear()
    self._key_handlers.clear()
    self._systems.clear()

  @property
  def space(self) -> 'Space | None':
    """The space to be used in system callbacks.

    A weak reference to the space is kept.
    The space may be None, possibly because of a dead weak reference.

    A space is required for processing events.
    A space is needed for _added_ and _removed_ callbacks, which
    are skipped if there is no space available.
    """
    return self._space()

  @space.setter
  def space(self, value: 'Space | None'):
    self._space = dead_weakref if value is None else weakref(value)

  @staticmethod
  def _insert_handler(
    lst: list[_EventHandler[_T]], handler: _EventHandler[_T], /
  ) -> None:
    """Insert a handler into a list of other handlers.

    Sorts by: highest priority, then oldest in manager.

    Args:
      lst: The list of handlers to add the handlers to.
      handler: The handler to be inserted into the list.
    """
    priority = handler.priority
    lo = 0
    hi = len(lst)
    while lo < hi:
      mid = (lo + hi) // 2
      if lst[mid].priority < priority:
        hi = mid
      else:
        lo = mid + 1
    lst.insert(lo, handler)

  class _SortKey:
    __slots__ = 'handler', 'systems'
    def __init__(self, handler: _EventHandler[Any], systems: Iterable[System], /):
      self.handler = handler
      self.systems = systems
    def __lt__(self, other: 'SystemManager._SortKey', /) -> bool:
      handler = self.handler
      other_handler = other.handler
      other_priority = other_handler.priority
      handler_priority = handler.priority
      if not (other_priority == handler_priority):
        return other_priority < handler_priority  # type: ignore[no-any-return]
      system = handler.system
      other_system = other_handler.system
      if system != other_system:
        for s in self.systems:
          if s == system:
            return True
          if s == other_system:
            return False
        raise ValueError
      name = handler.name
      other_name = other_handler.name
      if type(system) is not ModuleType:
        return name < other_name
      for n in system.__dict__:
        if other_name == n:
          return False
        if name == n:
          return True
      raise ValueError

  def _sort_handlers(
    self, handlers: Iterable[_EventHandler[_T]], /
  ) -> list[_EventHandler[_T]]:
    """Sort and return an iterable of handlers.

    Duplicate handlers are removed before sorting.

    Sorts by, in order:
    - highest priority
    - least recently added system
    - (same system) order the handlers were added in
      - if system is module instance, then order created
      - otherwise, alphabetical names

    Args:
      handlers: The iterable of event handlers to be sorted.

    Returns:
      A new list of sorted handlers from the original handlers.
    """
    SortKey = self._SortKey
    systems = self._systems
    # Uses dict to remove duplicates while preserving some order
    return sorted(dict.fromkeys(handlers), key=lambda h: SortKey(h, systems))

  def _get_handlers(self, event: _T, /) -> list[_EventHandler[_T]]:
    event_type = type(event)
    try:
      handlers = self._handlers[event_type]
    except KeyError:
      return []
    if event_type not in key_functions:
      return handlers[:]
    try:
      key_handlers = self._key_handlers[event_type]
    except KeyError:
      # A key function was added late
      self._key_handlers[event_type] = {}
      return handlers[:]
    key = key_functions[event_type](event)
    if not isinstance(key, Iterator):
      return (key_handlers.get(key, handlers))[:]
    keys = {k for k in key if k in key_handlers}
    if len(keys) > 1:
      return self._sort_handlers(
        [handler for key in keys for handler in key_handlers[key]]
      )
    return (key_handlers.get(keys.pop(), handlers) if keys else handlers)[:]

  @staticmethod
  def _get_bindings(system: System) -> list[tuple[Binding, _EventHandler[Any]]]:
    if type(system) is ModuleType:
      return [
        (binding, _EventHandler(system, binding._callback_, name, binding._priority_))
        for name, binding in system.__dict__.items() if isinstance(binding, Binding)
      ]
    return [
      (
        binding,
        _EventHandler(
          system,
          c if (c:=getattr(system, name)) is not binding else binding._callback_,
          name,
          binding._priority_
        )
      )
      for name in dict.fromkeys(dir(system))  # remove duplicates
      if isinstance(binding:=getattr_static(system, name), Binding)
    ]

  def _bind(self, system: System, /) -> list[EventHandlerAdded]:
    """Create handlers to process events for a system's bindings.

    Handlers of one event type are sorted by highest priority,
    then oldest system, then first one created in manager.

    Args:
      system: The system that was added, and to create handlers from.

    Returns:
      A list of EventHandlerAdded events to be posted.
    """
    all_handlers = self._handlers
    all_key_handlers = self._key_handlers
    insert_handler = self._insert_handler
    events: list[EventHandlerAdded] = []
    for binding, handler in self._get_bindings(system):
      event_type = binding._event_type_
      keys = binding._keys_
      if not keys:
        try:
          insert_handler(all_handlers[event_type], handler)
        except KeyError:
          all_handlers[event_type] = [handler]
        if event_type in key_functions:
          try:
            for handlers in all_key_handlers[event_type].values():
              insert_handler(handlers, handler)
          except KeyError:
            all_key_handlers[event_type] = {}
      else:
        try:
          handlers = all_handlers[event_type]
        except KeyError:
          handlers = all_handlers[event_type] = []
        if event_type not in all_key_handlers:
          handlers = handlers[:]
          insert_handler(handlers, handler)
          all_key_handlers[event_type] = {key: handlers[:] for key in keys}
        else:
          key_handlers = all_key_handlers[event_type]
          for key in keys:
            if key not in key_handlers:
              key_handlers[key] = handlers[:]
            insert_handler(key_handlers[key], handler)
      events.append(EventHandlerAdded(binding, handler))
    return events

  def _unbind(self, system: System, /) -> list[EventHandlerRemoved]:
    """Remove all handlers that belong to system from self.

    If an event type no longer has any handlers, it is removed.
    This also applies to keys with handlers.

    Args:
      system: The system that was removed, to remove handlers for.

    Returns:
      A list of EventHandlerRemoved events to be posted.
    """
    all_handlers = self._handlers
    all_key_handlers = self._key_handlers
    events: list[EventHandlerRemoved] = []
    for binding, handler in self._get_bindings(system):
      event_type = binding._event_type_
      handlers = all_handlers[event_type]
      handlers[:] = [
        handler for handler in handlers if handler.system != system
      ]
      if not handlers:
        del all_handlers[event_type]
      if event_type in all_key_handlers:
        key_handlers = all_key_handlers[event_type]
        for key, handlers in key_handlers.items():
          handlers[:] = [
            handler for handler in handlers if handler.system != system
          ]
          if not handlers:
            del key_handlers[key]
        if not key_handlers:
          del all_key_handlers[event_type]
      events.append(EventHandlerRemoved(binding, handler))
    return events
